import sys

import fuse

from tagfs.arguments import ArgumentParser, function_wrapper
from tagfs.ops import TagFS
from .db import Database


HELP_TEXT = """\
Usage: python -m tagfs FILE.tfs COMMAND ARGS...
Where command is:
  help, --help, -h        Show this message and exit
  mount DIR OPTIONS...    Mount tagfs to DIRectory
    -X                    Disable xattrs
    -A                    Disable __ALL__ tag
    -a                    Mount without tags (all files)
"""


@function_wrapper(name='mount.default')
def default(db: str, path: str) -> None:
    """Mount tagfs"""
    db = Database(db)
    fuse.FUSE(TagFS(db), path)


@function_wrapper
def mount(options: str, db: str, path: str) -> None:
    """Mount tagfs with options"""
    kw = {}

    for i in options.split(','):
        i = i.strip()
        n = i.find('=')
        if n == -1:
            kw[i] = True
        else:
            kw[i[n - 1:]] = i[:n]

    db = Database(db)
    fuse.FUSE(TagFS(db), path, raw_fi=False, **kw)


def main():
    p = ArgumentParser()

    p.add(('mount', 'default', ), default)
    p.add(('mount', ), mount)
    p.add_types(Database.CURSOR_TYPES, init_type=Database)
    if not p.parse(sys.argv):
        sys.exit(1)


if __name__ == '__main__':
    main()



import errno
import fuse
import os
import stat

from tagfs.selections import compile_selection
from .db import Database


class TagFS(fuse.Operations):
    OPTIONS = {
        'uid': int,
        'gid': int
    }

    def __init__(self, db: Database):
        self._db = db
        self._options = {}

    def __call__(self, name, *args, **kwargs):
        v = getattr(self, name, None)
        if not callable(v):
            raise fuse.FuseOSError(errno.EFAULT)
        return v(*args)

    def init(self, path):
        with self._db.cursor() as c:
            c = c.options

            for k, t in self.OPTIONS.items():
                v = c.get(k)
                if v is None:
                    v = self._corece(k, None)
                self._options[k] = t(v)

    @staticmethod
    def _corece(name, value):
        if name == 'uid':
            return os.geteuid() if value is None else int(value)
        if name == 'gid':
            return os.getegid() if value is None else int(value)
        return '' if value is None else value

    def getxattr(self, path, name, position=0):
        if path != '/' or not name.startswith('user.'):
            return b''
        return str(self._options.get(name[5:], '')).encode()

    def listxattr(self, path):
        if path != '/':
            return []
        return ['user.' + i for i in self._options.keys()]

    def removexattr(self, path, name):
        if path != '/' or not name.startswith('user.'):
            return

        name = name[5:]
        del self._options[name]
        self._db.cursor().options.unset(name)

    def setxattr(self, path, name, value, options, position=0):
        if path != '/' or not name.startswith('user.'):
            return

        name = name[5:]
        value = self._corece(name, value)
        self._options[name] = value
        self._db.cursor().options.set(name, value)

    def _make_attr(self, mode):
        return {'st_mode': mode, 'st_nlink': 2, 'st_uid': self._options['uid'], 'st_gid': self._options['gid']}

    def getattr(self, path, fh=None):
        if path == '/':
            return dict(st_mode=(stat.S_IFDIR | 0o755), st_nlink=2)

        _, *path = path.split('/')

        with self._db.cursor() as c:
            i = c.tags.get_id(path[0])
            if i is not None:
                if len(path) == 1:
                    return self._make_attr(stat.S_IFDIR | 0o755)

                if len(path) == 2:
                    j = c.files.get_id(path[1])
                    if j is not None and c.files.has_tag(j, i):
                        return self._make_attr(0o777 | stat.S_IFLNK)

            if c.selections.exists(path[0]):
                if len(path) == 1:
                    return self._make_attr(stat.S_IFDIR | 0o755)

                if len(path) == 2 and c.selections.matches(path[0], path[1]):
                    return self._make_attr(0o777 | stat.S_IFLNK)

        raise fuse.FuseOSError(errno.ENOENT)

    def readdir(self, path, fh):
        if path == '/':
            with self._db.cursor() as c:
                return ['.', '..', '__ALL__'] + c.tags.all_names() + c.selections.all_names()

        _, *path = path.split('/')
        if len(path) == 1:
            with self._db.cursor() as c:
                if c.tags.exists(path[0]):
                    return ['.', '..'] + c.files.get_by_tag(path[0])

                if c.selections.exists(path[0]):
                    return ['.', '..'] + c.selections.select(path[0])

        raise fuse.FuseOSError(errno.ENOENT)

    def mkdir(self, path, mode):
        if path == '/':
            raise fuse.FuseOSError(errno.EEXIST)

        _, *path = path.split('/')
        if len(path) != 1:
            raise fuse.FuseOSError(errno.EACCES)

        with self._db.cursor() as c:
            if not c.tags.new(path[0]):
                raise fuse.FuseOSError(errno.EEXIST)

    def rmdir(self, path):
        if path == '/':
            raise fuse.FuseOSError(errno.EINVAL)

        _, *path = path.split('/')
        if len(path) != 1:
            raise fuse.FuseOSError(errno.ENOTDIR)

        if path[0] == '__ALL__':
            raise fuse.FuseOSError(errno.EACCES)

        with self._db.cursor() as c:
            if not c.tags.remove(path[0]):
                raise fuse.FuseOSError(errno.ENOENT)

    def readlink(self, path):
        if path == '/':
            raise fuse.FuseOSError(errno.EINVAL)

        _, *path = path.split('/')
        if len(path) == 1:
            raise fuse.FuseOSError(errno.EINVAL)

        if len(path) != 2:
            raise fuse.FuseOSError(errno.ENOENT)

        with self._db.cursor() as c:
            return c.files.resolve(path[1])

    def unlink(self, path):
        if path == '/':
            raise fuse.FuseOSError(errno.EISDIR)

        _, *path = path.split('/')
        if len(path) == 1:
            raise fuse.FuseOSError(errno.EISDIR)

        if len(path) != 2:
            raise fuse.FuseOSError(errno.ENOENT)

        with self._db.cursor() as c:
            if not (c.files.remove(path[1]) if path[0] == '__ALL__' else c.files.remove_tag(path[1], path[0])):
                raise fuse.FuseOSError(errno.ENOENT)

    def symlink(self, path0, source):
        if path0 == '/':
            raise fuse.FuseOSError(errno.EEXIST)

        _, *path = path0.split('/')
        if len(path) == 1:
            with self._db.cursor() as c:
                if c.tags.exists(path[0]):
                    raise fuse.FuseOSError(errno.EEXIST)

        if source.startswith('sel:'):
            source = compile_selection(source)

            if len(path) != 1:
                raise fuse.FuseOSError(errno.EACCES)

            with self._db.cursor() as c:
                if not c.selections.new(path[0], source):
                    raise fuse.FuseOSError(errno.EEXIST)

        else:
            if len(path) == 1:
                raise fuse.FuseOSError(errno.EACCES)

            if len(path) != 2:
                raise fuse.FuseOSError(errno.ENOENT)

            source = os.path.normpath(os.path.join(path0, source))
            with self._db.cursor() as c:
                if not c.files.new(path[1], source):
                    raise fuse.FuseOSError(errno.EEXIST)

                if path[0] != '__ALL__' and not c.files.add_tag(path[1], path[0]):
                    raise fuse.FuseOSError(errno.EBUSY)

    def rename(self, old, new):
        if old == '/' or new == '/':
            raise fuse.FuseOSError(errno.EROFS)

        if not new.startswith('/'):
            new = os.path.normpath(os.path.join(old, new))

        _, *new = new.split('/')
        _, *old = old.split('/')

        with self._db.cursor() as c:
            if len(new) == len(old) == 1 and not c.tags.rename(old[0], new[0]):
                raise fuse.FuseOSError(errno.EROFS)

            elif len(new) == len(old) == 2:
                if new[0] != old[0]:
                    if old[0] != '__ALL__':
                        c.files.remove_tag(old[1], old[0])

                    if new[0] != '__ALL__':
                        c.files.add_tag(old[1], new[0])

                if new[1] != old[1]:
                    c.files.rename(old[1], new[1])



import functools
import sqlite3
import threading
import typing

name_id = typing.Union[str, int]


class Database(threading.local):
    INIT = """
    CREATE TABLE IF NOT EXISTS tags(id INTEGER PRIMARY KEY, name TEXT UNIQUE);
    CREATE TABLE IF NOT EXISTS tag_files(tag_id INTEGER, file_id INTEGER);
    CREATE TABLE IF NOT EXISTS files(id INTEGER PRIMARY KEY, name TEXT, path TEXT);
    CREATE TABLE IF NOT EXISTS options(name TEXT UNIQUE, value);
    CREATE TABLE IF NOT EXISTS selections(name TEXT UNIQUE, value TEXT);
    """

    CURSOR_TYPES = {}

    def __init__(self, db, *args, **kwargs):
        self._db = sqlite3.connect(db, *args, **kwargs)
        self._db.executescript(self.INIT)
        self._db.commit()

    def cursor(self):
        return _BaseCursor(self._db)

    def __getattr__(self, item):
        c = self.cursor()
        v = getattr(c, item)

        if not callable(v):
            return v

        @functools.wraps(v)
        def wr(*args, **kwargs):
            with c:
                return v(*args, **kwargs)

        return wr


class _BaseCursor:
    def __init__(self, db):
        if isinstance(db, _BaseCursor):
            self._db = db._db
            self._c = db._c

        else:
            self._db = db
            self._c = db.cursor()

        self.execute = self._c.execute
        self.fetchone = self._c.fetchone
        self.fetchall = self._c.fetchall
        self.fetchmany = self._c.fetchmany
        self.executescript = self._c.executescript
        self.__iter__ = self._c.__iter__
        self.close = self._c.close
        self.commit = self._db.commit

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()

    def __init_subclass__(cls, **kwargs):
        Database.CURSOR_TYPES[cls.ATTR_NAME] = cls

    def _fetch_first(self):
        return [i[0] for i in self.fetchall()]

    def __getattr__(self, item):
        if item == 'fetch_first':
            return self._fetch_first

        v = Database.CURSOR_TYPES.get(item)
        if v is None:
            raise AttributeError('%r object has no attribute %r' % (self.__class__.__name__, item))
        return v(self)


class _SelectionsCursor(_BaseCursor):
    ATTR_NAME = 'selections'

    def new(self, name: str, value: str):
        """
        Create new selection
        :param name:
        :param value:
        :return:
        """
        self.execute("INSERT INTO selections(name, value) VALUES(?, ?)", (name, value))
        self.commit()
        return True

    def remove(self, name: str):
        """
        Remove selection by name or id
        :param name:
        :return:
        """

        self.execute("DELETE FROM selections WHERE name = ?", (name, ))
        self.commit()
        return True

    def all_names(self):
        """
        Show all selection names
        :return:
        """
        self.execute("SELECT name FROM selections;")
        return self.fetch_first()

    def rename(self, old: str, new: str):
        """
        Change name of tag from old (may be id) to new
        :param old:
        :param new:
        :return:
        """
        if '__ALL__' in (old, new):
            return False

        self.execute("UPDATE selections SET name = ? WHERE name = ?", (new, old))
        self.commit()
        return True

    def resolve(self, name: str):
        """
        Return value of selection
        :param name:
        :return:
        """
        self.execute("SELECT value FROM selections WHERE name = ?", (name, ))
        c = self.fetchone()
        return None if c is None else c[0]

    def select(self, name: str):
        """
        Return all elements matches selection
        :param name:
        :return:
        """
        self.execute("SELECT files.name FROM tag_files INNER JOIN files ON files.id = tag_files.file_id INNER JOIN tags"
                     " ON tags.id = tag_files.tag_id WHERE %s" % self.resolve(name))
        return self.fetch_first()

    def exists(self, name: str):
        """
        Return true if selection exists
        :param name:
        :return:
        """
        return self.resolve(name) is not None

    def matches(self, name: str, file: name_id):
        """
        Return true if selection contains file
        :param name:
        :param file:
        :return:
        """
        if isinstance(file, int):
            self.execute("SELECT files.name FROM tag_files INNER JOIN files ON files.id = tag_files.file_id INNER JOIN "
                         "tags ON tags.id = tag_files.tag_id WHERE files.id = ? AND %s" % self.resolve(name), (file, ))
        else:
            self.execute("SELECT files.name FROM tag_files INNER JOIN files ON files.id = tag_files.file_id INNER JOIN "
                         "tags ON tags.id = tag_files.tag_id WHERE files.name = ? AND %s" % self.resolve(name), (file,))
        c = self.fetchone()
        return c is not None


class _OptionsCursor(_BaseCursor):
    ATTR_NAME = 'options'

    def set(self, name: str, value):
        """
        Set filesystem option
        :param name:
        :param value:
        :return:
        """
        self.execute("INSERT OR REPLACE INTO options(name, value) VALUES(?, ?)", (name, value))
        self.commit()

    def get(self, name: str):
        """
        Get filesystem option
        :param name:
        :return:
        """
        self.execute("SELECT value FROM options WHERE name = ?", (name, ))
        c = self.fetchone()
        return None if c is None else c[0]

    def unset(self, name: str):
        """
        Remove filesystem option
        :param name:
        :return:
        """
        self.execute("REMOVE FROM options WHERE name = ?", (name, ))
        self.commit()


class _TagsCursor(_BaseCursor):
    ATTR_NAME = 'tags'

    def new(self, name: str):
        """
        Create new tag
        :param name:
        :return:
        """
        if name == '__ALL__':
            return False

        self.execute("INSERT INTO tags(name) VALUES (?)", (name, ))
        self.commit()
        return True

    def remove(self, name: name_id):
        """
        Remove tag by name or id
        :param name:
        :return:
        """
        if name == '__ALL__':
            return False

        i = self.get_id(name)

        self.execute("DELETE FROM tag_files WHERE tag_id = ?", (i, ))
        self.execute("DELETE FROM tags WHERE id = ?", (i, ))
        self.commit()
        return True

    def get_id(self, name: name_id):
        """
        Return tag id by name (or id)
        :param name:
        :return:
        """
        if isinstance(name, int):
            return name
        if name == '__ALL__':
            return -1

        self.execute("SELECT id FROM tags WHERE name = ?", (name,))
        c = self.fetchone()
        return None if c is None else c[0]

    def get_name(self, name: name_id):
        """
        Return name of tag by id (or name)
        :param name:
        :return:
        """
        if isinstance(name, str):
            return name

        self.execute("SELECT name FROM tags WHERE id = ?", (name,))
        c = self.fetchone()
        return None if c is None else c[0]

    def all_names(self):
        """
        Show all tag names
        :return:
        """
        self.execute("SELECT name FROM tags;")
        return self.fetch_first()

    def all_ids(self):
        """
        Show all tag ids
        :return:
        """
        self.execute("SELECT id FROM tags;")
        return self.fetch_first()

    def rename(self, old: name_id, new: str):
        """
        Change name of tag from old (may be id) to new
        :param old:
        :param new:
        :return:
        """
        if '__ALL__' in (old, new):
            return False

        if isinstance(old, int):
            self.execute("UPDATE tags SET name = ? WHERE id = ?", (new, old))
        else:
            self.execute("UPDATE tags SET name = ? WHERE name = ?", (new, old))

        self.commit()
        return True

    def exists(self, name: name_id):
        """
        Return true if tag exists
        :param name:
        :return:
        """
        return self.get_id(name) is not None


class _FilesCursor(_BaseCursor):
    ATTR_NAME = 'files'

    def new(self, name: str, path: str):
        """
        Create new file link
        :param name:
        :param path:
        :return:
        """
        self.execute("INSERT INTO files(name, path) VALUES (?, ?)", (name, path))
        self.commit()
        return True

    def remove(self, name: name_id):
        """
        Remove file link by name or id
        :param name:
        :return:
        """
        i = self.get_id(name)

        self.execute("DELETE FROM tag_files WHERE file_id = ?", (i, ))
        self.execute("DELETE FROM files WHERE id = ?", (i, ))
        self.commit()
        return True

    def get_id(self, name: name_id):
        """
        Return id of file by name (or id)
        :param name:
        :return:
        """
        if isinstance(name, int):
            return name

        self.execute("SELECT id FROM files WHERE name = ?", (name,))
        c = self.fetchone()
        return None if c is None else c[0]

    def get_name(self, name: name_id):
        """
        Return name of file by id (or name)
        :param name:
        :return:
        """
        if isinstance(name, str):
            return name

        self.execute("SELECT name FROM files WHERE id = ?", (name,))
        c = self.fetchone()
        return None if c is None else c[0]

    def all_names(self):
        """
        Return all files names
        :return:
        """
        self.execute("SELECT name FROM files;")
        return self.fetch_first()

    def all_ids(self):
        """
        Return all files ids
        :return:
        """
        self.execute("SELECT id FROM files;")
        return self.fetch_first()

    def rename(self, old: name_id, new: str):
        """
        Rename file from old (may be id) to new
        :param old:
        :param new:
        :return:
        """
        if isinstance(old, int):
            self.execute("UPDATE files SET name = ? WHERE id = ?", (new, old))
        else:
            self.execute("UPDATE files SET name = ? WHERE name = ?", (new, old))

        self.commit()
        return True

    def exists(self, name: name_id):
        """
        Return true if exists
        :param name:
        :return:
        """
        return self.get_id(name) is not None

    def get_by_tag(self, tag: name_id):
        """
        Return all file name with given tag (name or id)
        :param tag:
        :return:
        """
        if tag == '__ALL__':
            return self.all_names()

        t = self.tags.get_id(tag)
        self.execute("SELECT files.name FROM files INNER JOIN tag_files ON tag_files.file_id = files.id WHERE "
                     "tag_files.tag_id = ?", (t, ))
        return self.fetch_first()

    def add_tag(self, name: name_id, tag: name_id):
        """
        Add tag to file
        :param name:
        :param tag:
        :return:
        """
        if tag == '__ALL__':
            return False

        self.execute("INSERT INTO tag_files(tag_id, file_id) VALUES(?, ?)",
                     (self.tags.get_id(tag), self.get_id(name)))
        self.commit()
        return True

    def remove_tag(self, name: name_id, tag: name_id):
        """
        Remove tag from file
        :param name:
        :param tag:
        :return:
        """
        if tag == '__ALL__':
            return False

        self.execute("DELETE FROM tag_files WHERE tag_id = ? AND file_id = ?",
                     (self.tags.get_id(tag), self.get_id(name)))
        self.commit()
        return True

    def has_tag(self, name: name_id, tag: name_id):
        """
        Return true if file has tag
        :param name:
        :param tag:
        :return:
        """
        if tag in ('__ALL__', -1):
            return True

        self.execute("SELECT tag_id FROM tag_files WHERE tag_id = ? AND file_id = ?",
                     (self.tags.get_id(tag), self.get_id(name)))
        return len(self.fetch_first()) > 0

    def resolve(self, name: name_id):
        """
        Return path of file
        :param name:
        :return:
        """
        if isinstance(name, int):
            self.execute("SELECT path FROM files WHERE id = ?", (name, ))
        else:
            self.execute("SELECT path FROM files WHERE name = ?", (name, ))

        c = self.fetchone()
        return None if c is None else c[0]

    def get_tags(self, name: name_id):
        """
        Return all tags of file
        :param name:
        :return:
        """

        name = self.get_id(name)
        self.execute("SELECT tags.name FROM tag_files INNER JOIN tags ON tags.id = tag_files.tag_id WHERE"
                     " tag_files.file_id = ?", (name, ))
        return self.fetch_first()

    def set_tags(self, name: name_id, tags: typing.List[name_id]):
        """
        Replace tags of file
        :param name:
        :param tags:
        :return:
        """

        name = self.get_id(name)
        self.execute("DELETE FROM tag_files WHERE file_id = ?", (name, ))

        t = self.tags
        for tag in tags:
            i = t.get_id(tag)
            if i is None:
                t.new()
                i = t.get_id(tag)

            self.execute("INSERT INTO tag_files(tag_id, file_id) VALUES(?, ?)", (i, name))

        self.commit()
        return True



import builtins
import functools
import inspect
import sys
import typing


class ArgumentParser:
    FIRST_LINES = (
        'Usage: {program} COMMAND ARGS...',
        'Where command is:'
    )

    def __init__(self, *, types=None, help_text='', add_first_line=True, init_type=None):
        self._help = help_text.splitlines(False)
        self._args = []

        if add_first_line:
            self._help = list(self.FIRST_LINES) + self._help

        if types is not None:
            self.add_types(types, init_type=init_type)

    def add(self, pattern, typ, *, init_type=None):
        if init_type is None:
            init_type = typ

        obj = getattr(typ, pattern[-1])
        new = self._get_type_init(init_type)
        self._args.append((pattern, init_type, new, obj))
        x = ' '.join((new.__code__.co_varnames[i + 1].upper() for i in range(new.__code__.co_argcount - 1))) + ' '\
            + ' '.join(pattern) + ' '\
            + ' '.join((obj.__code__.co_varnames[i + 1].upper() for i in range(obj.__code__.co_argcount - 1)))
        self._help.append('  {:<23.23} {}'.format(x, self._get_doc(obj)))

    def add_type(self, typ, *, prefix=(), init_type=None):
        for mn in dir(typ):
            if not mn.startswith('_') and callable(getattr(typ, mn)):
                self.add((*prefix, mn), typ, init_type=init_type)

    def add_types(self, types, *, prefix=(), init_type=None):
        for name, typ in types.items():
            self.add_type(typ, prefix=prefix + (name, ), init_type=init_type)

    def print_help(self, program, error=False):
        v = '\n'.join(self._help).format(program=program)
        if error:
            print(v, file=sys.stderr)
        else:
            print(v)

    def parse(self, args):
        args = list(args)
        program = args.pop(0)

        if not args:
            self.print_help(program, True)
            return False

        for pattern, typ, new, obj in self._args:
            bl = new.__code__.co_argcount - 1
            al = obj.__code__.co_argcount - 1

            if len(pattern) + al + bl != len(args):
                continue

            for i, v in enumerate(pattern):
                if args[i + bl] != v:
                    break
            else:
                argv = []
                lp = len(pattern) + bl

                for i in range(bl):
                    argv.append(self._convert_arg(new, i, args[i]))

                v = typ(*argv)
                argv = []

                for i in range(al):
                    argv.append(self._convert_arg(obj, i, args[i + lp]))

                v = self._getattr_long(v, *pattern)(*argv)
                if v is not None:
                    print(v)
                return True

        self.print_help(program, True)
        return False

    @staticmethod
    def _getattr_long(obj, *args):
        for i in args:
            obj = getattr(obj, i)
        return obj

    @staticmethod
    def _get_doc(v):
        v = getattr(v, '__doc__', None)
        if v is not None:
            v = v.strip()

        if not v:
            return 'Undocumented'
        return v.splitlines()[0]

    @staticmethod
    def _convert_arg(method, arg_no, val: str, exc_type=None):
        start = 0
        i = -1

        while True:
            li = i
            i = val.find(':', start)
            if i == -1:
                break

            if i + 1 != len(val) and val[i + 1] == ':':
                val = val[:i] + val[i + 1:]
                start = i + 1
                i = -1
            else:
                start = i + 1

        if li != -1:
            return getattr(builtins, val[li + 1:])(val[:li])

        if exc_type is None:
            typ = method.__annotations__.get(method.__code__.co_varnames[arg_no + 1])
            if typ is None:
                return val

        else:
            typ = exc_type

        # noinspection PyUnresolvedReferences,PyProtectedMember
        if isinstance(typ, typing._Union):
            for i in typ.__args__:
                try:
                    return i(val)
                except TypeError:
                    pass
                except ValueError:
                    pass

        elif issubclass(typ, typing.List):
            # noinspection PyUnresolvedReferences
            item_type = typ.__args__[0]
            return [ArgumentParser._convert_arg(method, arg_no, item, item_type)
                    for item in val.split(',')]

        if callable(typ):
            try:
                return typ(val)
            except TypeError:
                pass
            except ValueError:
                pass
        return val

    @staticmethod
    def _get_type_init(typ):
        i = getattr(typ, '__init__', None)
        if hasattr(i, '__code__'):
            return i
        i = getattr(typ, '__new__', None)
        if hasattr(i, '__code__'):
            return i
        raise TypeError("'__init__' and '__new__' for type %r is native!" % typ.__name__)


def function_wrapper(func=None, *, name=None):
    if func is not None:
        return _function_wrapper(func.__name__, func)
    return functools.partial(_function_wrapper, name)


def _function_wrapper(name, func):
    name = name.split('.')
    mod = func.__module__
    init = _init_none
    func = func0 = _make_fn(func, name[-1], mod)

    if len(name) == 1:
        return type(name[0], (), {
            name[0]: func,
            '__module__': mod,
            '__init__': init
        })

    for i in reversed(name[1:]):
        func = type(i, (), {
            i: func,
            '__module__': mod,
            '__init__': init
        })()

    return type(name[0], (), {
        name[0]: func,
        name[-1]: func0,
        '__module__': mod,
        '__init__': init
    })


def _init_none(self):
    pass


def _make_fn(fn, name, mod):
    s = inspect.signature(fn, follow_wrapped=False)
    pn = []

    for i, v in s.parameters.items():
        if v.kind is inspect.Parameter.KEYWORD_ONLY:
            if v.default is inspect.Parameter.empty:
                raise ValueError("Can't wrap function with non-default keyword-only arguments")
            i = '{0}={0}'.format(i)
        pn.append(i)

    c = 'def %s(self, %s:\n    return __fn__(%s)' % (name, str(s)[1:], ', '.join(pn))
    g = {'__fn__': fn}
    exec(c, g, g)
    f = g[name]
    f.__doc__ = fn.__doc__
    f.__module__ = mod
    return f



import ast


_OPS = {
    ast.Eq: '=',
    ast.NotEq: '!=',
    ast.Gt: '>',
    ast.GtE: '>=',
    ast.Lt: '<',
    ast.LtE: '<=',
    ast.Is: '=',
    ast.IsNot: '!=',
    ast.In: 'IN',
    ast.NotIn: 'NI',
    ast.And: 'AND',
    ast.Or: 'OR',
    ast.Not: '!',
}

_NAMES = {
    'file': {
        'name': 'files.name',
        'path': 'files.path',
        'id': 'files.id',
    },
    'tag': {
        'name': 'tags.name',
        'id': 'tags.id'
    }
}


class SelectionBuilder:
    def __init__(self):
        self.result = self._curr = {}

    def _fix_in(self, op, val):
        if op is ast.In or op is ast.NotIn:
            op = ast.Eq if op is ast.In else ast.NotEq
            val = val.elts
        return op, val

    def _build_cmp(self, cmp: ast.Compare):
        v = []
        p = cmp.left

        for op, val in zip(cmp.ops, cmp.comparators):
            op = type(op)
            if op is ast.In or op is ast.NotIn:
                r = [self._build(i) for i in val.elts]

            else:
                r = [self._build(val)]

            v.append({
                'op': _OPS[op],
                'values': [self._build(p)] + r
            })
            p = val

        if len(v) == 1:
            return v[0]

        return {
            'op': 'AND',
            'values': v
        }

    def _build_bool_op(self, cmp: ast.BoolOp):
        return {
            'op': _OPS[type(cmp.op)],
            'values': [self._build(i) for i in cmp.values]
        }

    def _build_str(self, val: ast.Str):
        return {
            'type': 'str',
            'value': val.s,
        }

    def _build_num(self, val: ast.Num):
        return {
            'type': 'number',
            'value': val.n,
        }

    def _build_attr(self, val: ast.Attribute):
        return _NAMES[val.value.id][val.attr]

    def _build(self, val):
        if isinstance(val, ast.BoolOp):
            return self._build_bool_op(val)

        if isinstance(val, ast.Compare):
            return self._build_cmp(val)

        if isinstance(val, ast.Attribute):
            return self._build_attr(val)

        if isinstance(val, ast.Str):
            return self._build_str(val)

        if isinstance(val, ast.Num):
            return self._build_num(val)

        raise ValueError("Unknown or invalid expression")

    def build(self, value):
        if not isinstance(value, ast.Expression):
            raise TypeError("Value must be expression")

        return self._build(value.body)

    def compile(self, v):
        if 'op' in v:
            op = v['op']
            if op == 'IN':
                a = self.compile(v['values'][0])
                return '(%s)' % 'OR'.join(('(%s=%s)' % (a, self.compile(i)) for i in v['values'][1:]))

            if op == 'NI':
                a = self.compile(v['values'][0])
                return '(%s)' % 'AND'.join(('(%s!=%s)' % (a, self.compile(i)) for i in v['values'][1:]))

            return '(%s)' % op.join((self.compile(i) for i in v['values']))
        if 'type' in v:
            return repr(v['value'])
        return v


def compile_selection(sel: str):
    if sel.startswith('sel:'):
        sel = sel[4:]

    b = SelectionBuilder()
    sel = compile(sel, '<selection>', 'eval', ast.PyCF_ONLY_AST, 1)
    return b.compile(b.build(sel))



#!/usr/bin/python3.6

from setuptools import setup, find_packages

setup(
    name='tagfs',
    author='Pasha__kun',
    version='1.0.1',
    license='MIT',
    packages=find_packages(),
    entry_points={
        'console_scripts': [
            'tagfs = tagfs.__main__:main'
        ]
    }
)
